(* digit_to_z : this function is supposed to be used to map integers
  to numbers of type Z.  The basic functions on Z are passed as arguments.
  This function is to be used only on non-negative numbers. *)


let rec digit_to_z (zero : 'a) 
  (one : 'a) (twicep1 : 'a -> 'a) (twice : 'a -> 'a) (x : int) = 
  if x = 0 then zero
  else if x = 1 then one
  else if x mod 2 = 1 then twicep1 (digit_to_z zero one twicep1 twice (x / 2))
  else twice (digit_to_z zero one twicep1 twice (x / 2))

(* the number 10. *)
let z10 twicep1 twice one = twice (twicep1 (twice one))

(* z_of_string : string -> z
  This function should accept strings of arbitrary length, containing only
  digits, with possible a '-' character as first character (this what
  the mutable started is used for.) *)
let z_of_string zero one twicep1 twice (zmult : 'a -> 'a -> 'a)
  (zplus : 'a -> 'a -> 'a) (zopp : 'a -> 'a) (s : string) : 'a =
  let sign = if String.get s 0 = '-' then zopp one else one in
  let v = ref zero in
  let started = ref false in
  (String.iter 
    (fun c ->
       (if Char.code c <= Char.code '9' && Char.code '0' <= Char.code c then
          v := 
          zplus (zmult (z10 twicep1 twice one) !v)
            (digit_to_z zero one twicep1 twice (Char.code c - Char.code '0'))
       else if c = '-' && !started then ()
       else  failwith ("unexpected character in a number : " ^ (String.make 1 c)
                       ^ "\n")); started := true) s; zmult sign !v)

let rec z_to_int 
  (pattern_match_number :
     'unit -> ('unit -> 'a) -> ('unit -> 'a) ->
     ('b -> 'a) ->
     ('b -> 'a) -> ('b -> 'a) -> 'b -> 'a) (num : 'b) : int =
  pattern_match_number ()
    (* num=0 *) (fun _ -> 0)
    (* num=1 *) (fun _ -> 1)
    (* num <0, num = -p *) (fun p -> - z_to_int pattern_match_number p)
    (* num=2*p+1 *) (fun p -> 2 * z_to_int pattern_match_number p + 1 )
    (* num=2*p *) (fun p -> 2 * z_to_int pattern_match_number p) num


let rec print_z_rec   (pattern_match_number_unit:
    unit -> (unit -> unit) -> (unit -> unit) -> 
    ('z -> unit) -> ('z -> unit) -> ('z -> unit) -> 'z -> unit)
       pattern_match_number_z
  (div10_cont : ('z -> 'z -> 'a) -> 'z -> 'a) num =
  let non_null_case n =
    div10_cont (fun q r -> print_z_rec pattern_match_number_unit
         pattern_match_number_z div10_cont q;
    print_int (z_to_int pattern_match_number_z r)) n in
  pattern_match_number_unit ()
    (* num=0 *) (fun _ -> ())
    (* num=1 *) (fun _ -> non_null_case num)
    (* num < 0, num=-p *) (fun p -> assert false)
    (* num=2*p+1 *) (fun _ -> non_null_case num)
    (* num=2*p *) (fun _ -> non_null_case num)
    num

(* print_z. this function is to be used to print a number in decimal form.
  it is parameterized by a pattern matching function and function for
  dividing by 10. The type of these function is chosen to be independent
  from any choice data-types by the extractor. *)

let print_z 
  (pattern_match_number_unit:
    unit -> (unit -> unit) -> (unit -> unit) -> 
    ('z -> unit) -> ('z -> unit) -> ('z -> unit) -> 'z -> unit)
  pattern_match_number_z div10_cont n =
  pattern_match_number_unit ()
    (* num=0 *) (fun _ -> print_int 0)
    (* num=1 *) (fun _ -> print_int 1)
    (* num < 0, num=-p *) (fun p -> print_string "-";
         print_z_rec pattern_match_number_unit pattern_match_number_z 
         div10_cont p)
    (* num=2*p+1 *) (fun _ -> print_z_rec pattern_match_number_unit 
         pattern_match_number_z div10_cont n)
    (* num=2*p *) (fun _ -> print_z_rec pattern_match_number_unit
         pattern_match_number_z div10_cont n)
    n

let array_to_list list_cons list_nil a =
  let len = Array.length a in
  let rec al n = if n = len then list_nil else list_cons a.(n) (al (n+1)) in
    al 0;;

(* native_to_string transforms an Ocaml string into a coq string.
  it relies on digit_to_Z. *)

let native_to_string zero one twicep1 twice string_cons empty s =
  let len = String.length s in
  let rec ns n = 
    if n = len then empty
    else
     string_cons 
       (digit_to_z zero one twicep1 twice (Char.code (String.get s n)))
       (ns (n+1)) in
     ns 0
   
let input_to_string () =
  let rec read_all sl =
    let opt_s = try Some (read_line ()) with End_of_file -> None in
    match opt_s with None -> List.rev sl | Some s -> read_all (s::sl) in
  String.concat "\n" (read_all [])

let output_char (n:int) = (output_char stdout (Char.chr n); true)

let flush_stdout () = flush stdout
    
   
